import numpy as np
import mbam.finite_difference as fd
from mbam.utils import initial_velocity

from model_linear import M, N, f, J, xi


_, _, vh = np.linalg.svd(J)
jac_fn = fd.jacobian_func(f, M, N)
Avv_fn = fd.Avv_func(f)


def test_initial_velocity():
    """We test the utility function to get initial velocity using a linear
    model. Since Avv for a linear model is zero, then the initial velocities
    should be parallel to the right singular vectors of the Jacobian. The
    direction might be different, though, since the ``initial_velocity``
    returns the direction in which the velocity increases.
    """
    for i in range(N):
        for forward in [True, False]:
            v = initial_velocity(xi, jac_fn, Avv_fn, i=i, forward=forward)
            v_ref = vh[-1 - i] * (int(forward) * 2 - 1)
            assert np.isclose(np.abs(v @ v_ref), 1.0)


if __name__ == "__main__":
    test_initial_velocity()



"""An example of calculating a geodesic. Model takes the form

.. math::

    y(t,x) = e^{-x_0 t} + e^{-x_1 t},

for time points :math:`t = \{0.5, 1.0, 2.0\}`.

We enforce :math:`x_i > 0` by going to log parameters:

.. math::

    y(t,x) = e^{-\exp(x_0) t} + e^{-\exp(x_1) t}.

We adopt the convention that the model has N parameters and makes M
predictions. Then, the output of :math:`r(x)` should be a vector length M the
output of :math:`j(x)` (i.e., jacobian) should be an :math:`M \times N` matrix.
The output of :math:`Avv(x,v)` should be a vector of length M. In this example,
:math:`M = 3` (three time points) and :math:`N = 2`.
"""

from mbam import Geodesic, initial_velocity
import numpy as np
import matplotlib.pyplot as plt


exp = np.exp

t = np.array([0.5, 1.0, 2.0])


# Model predictions
def r(x):
    return exp(-exp(x[0]) * t) + exp(-exp(x[1]) * t)


# Jacobian
def j(x):
    return np.array(
        [
            -t * exp(x[0]) * exp(-exp(x[0]) * t),
            -t * exp(x[1]) * exp(-exp(x[1]) * t),
        ]
    ).T


# Directional second derivative
def Avv(x, v):
    h = 1e-4
    return (r(x + h * v) + r(x - h * v) - 2 * r(x)) / h / h


# Choose starting parameters
x = np.log([1.0, 2.0])
v = initial_velocity(x, j, Avv)


# Callback function used to monitor the geodesic after each step
def callback(g):
    # Integrate until the norm of the velocity has grown by a factor of 10
    # and print out some diagnotistic along the way
    print(
        "Iteration: %i, tau: %f, |v| = %f"
        % (len(g.vs), g.ts[-1], np.linalg.norm(g.vs[-1]))
    )
    return np.linalg.norm(g.vs[-1]) < 20.0


# Construct the geodesic
# It is usually not necessary to be very accurate here, so we set small
# tolerances
geo = Geodesic(r, j, Avv, x, v, atol=1e-2, rtol=1e-2, callback=callback)

# Integrate
geo.integrate(25.0)

# Plot the geodesic path to find the limit
# This should show the singularity at the "fold line" x[0] = x[1]
plt.figure()
plt.plot(geo.ts, geo.xs)
plt.xlabel("tau")
plt.ylabel("Parameter Values")
plt.show()



import numpy as np

try:
    import jax.numpy as jnp

    jax_avail = True
except ModuleNotFoundError:
    jax_avail = False


np.random.seed(1717)

# Time series
t = np.array([1.0, 2.0, 3.0])
if jax_avail:
    t_ad = jnp.array(t)  # For JAX

# Dimensionality
M = len(t)
N = 2


# Define the model
def f(x):
    """A function that computes the model predictions."""
    return np.exp(-x[0] * t) + np.exp(-x[1] * t)


if jax_avail:

    def f_ad(x):
        """Also a function to compute the predictions, but written for using with
        JAX.
        """
        return jnp.exp(-x[0] * t_ad) + jnp.exp(-x[1] * t_ad)


def jac_an(x):
    """Compute the jacobian analytically. -t*exp(-th*t)"""
    return np.array([-t * np.exp(-x[0] * t), -t * np.exp(-x[1] * t)]).T


def Avv_an(x, v):
    """Compute Avv analytically.t^2 * exp(-th*t)"""
    H = np.array(
        [
            [t ** 2 * np.exp(-x[0] * t), np.zeros(M)],
            [np.zeros(M), t ** 2 * np.exp(-x[1] * t)],
        ]
    )
    H = np.moveaxis(H, (0, 1, 2), (1, 2, 0))
    return H @ v @ v


# Initial conditions
xi = np.exp(np.random.randn(N))
vi = np.random.randn(N)



"""Contains functions to compute the derivative of the model with finite
difference formula.
"""

import numpy as np


# Forward Difference Formulas
def FD(f, x, v, h):
    """2-point forward difference"""
    return (f(x + h * v) - f(x)) / h


def FD2(f, x, v, h):
    """3-point forward difference"""
    return (-f(x + 2 * h * v) + 4 * f(x + h * v) - 3 * f(x)) / (2 * h)


def FD3(f, x, v, h):
    """4-point forward difference"""
    return (
        f(x + 4 * h * v) - 12 * f(x + 2 * h * v) + 32 * f(x + h * v) - 21 * f(x)
    ) / (12 * h)


def FD4(f, x, v, h):
    """5-point forward difference"""
    return (
        -f(x + 8 * h * v)
        + 28 * f(x + 4 * h * v)
        - 224 * f(x + 2 * h * v)
        + 512 * f(x + h * v)
        - 315 * f(x)
    ) / (168 * h)


# Center Difference Formulas
def CD(f, x, v, h):
    """2-point central difference"""
    return (f(x + h * v) - f(x - h * v)) / (2 * h)


def CD4(f, x, v, h):
    """4-point central difference"""
    return (
        -f(x + 2 * h * v)
        + 8 * f(x + h * v)
        - 8 * f(x - h * v)
        + f(x - 2 * h * v)
    ) / (12 * h)


# Avv Formulas
def AvvCD(f, x, v, h):
    """Directional second derivative Avv with 2-point central difference"""
    return (f(x + h * v) + f(x - h * v) - 2 * f(x)) / (h ** 2)


def AvvCD4(f, x, v, h):
    """Directional second derivative Avv with 4-point central difference"""
    return (
        -f(x + 2 * h * v)
        + 16 * f(x + h * v)
        + 16 * f(x - h * v)
        - f(x - 2 * h * v)
        - 30 * f(x)
    ) / (12 * h ** 2)


# Auv formulas
def AuvCD(f, x, u, v, h):
    """Directional second derivative Auv with 2-point central difference"""
    return (
        f(x + h * u + h * v)
        - f(x + h * u - h * v)
        - f(x - h * u + h * v)
        + f(x - h * u - h * v)
    ) / (4 * h ** 2)


def AuvCD4(f, x, u, v, h):
    """Directional second derivative Auv with 4-point central difference"""
    return (
        -f(x + 2 * h * u + 2 * h * v)
        + f(x + 2 * h * u - 2 * h * v)
        + 16 * f(x + h * u + h * v)
        - 16 * f(x + h * u - h * v)
        - 16 * f(x - h * u + h * v)
        + 16 * f(x - h * u - h * v)
        + f(x - 2 * h * u + 2 * h * v)
        - f(x - 2 * h * u - 2 * h * v)
    ) / (48 * h ** 2)


################################################################################
# Wrapper functions to convert the finite difference formulae above to the
# functions required by Geodesic.


def jacobian_func(f, M, N, deriv_fn=FD, h=0.1):
    """A wrapper function convert functions to compute directional first
    derivative to the jacobian function needed in :class:`~mbam.geodesic`.

    Parameters
    ----------
    f: callable ``f(x)``
        Function to evaluate the model.
    M: int
        Number of predictions the model makes.
    N: int
        Number of parameters the model takes.
    deriv_fn: callable ``deriv_fn(f, x, v, h)`` (optional)
        A function to compute the directional derivative of the model.
    h: float (optional)
        Finite difference step size.

    Returns
    -------
    jacobian: callable ``jacobian(x)``
        A function that evaluates the Jacobian of the model ``f`` at parameter
        ``x``.

    Notes
    -----
    If the predictions are weighted differently, the argument ``f`` should be
    the residual function, assuming that the weights are encoded in the
    residual.
    """

    def jacobian(x):
        # This matrix will be used to specify the direction of the derivative
        eye = np.diag(np.ones(N))
        # Compute the jacobian matrix
        jac = np.empty((M, N))
        for ii, v in enumerate(eye):
            jac[:, ii] = deriv_fn(f, x, v, h)
        return jac

    return jacobian


def Avv_func(f, Avv_fn=AvvCD, h=0.1):
    """A wrapper function convert functions to compute directional second
    derivative to the Avv function needed in :class:`~mbam.geodesic`.

    Parameters
    ----------
    f: callable ``f(x)``
        Function to evaluate the model.
    Avv_fn: callable ``Avv_fn(f, x, v, h)`` (optional)
        A function to compute the directional second derivative of the model.
    h: float (optional)
        Finite difference step size.

    Returns
    -------
    Avv: callable ``Avv(x, v)``
        A function that evaluates the directional second derivative (Avv) of
        the model ``r`` at parameter ``x`` in the ``v`` direction.

    Notes
    -----
    If the predictions are weighted differently, the argument ``f`` should be
    the residual function, assuming that the weights are encoded in the
    residual.
    """

    def Avv(x, v):
        return Avv_fn(f, x, v, h)

    return Avv



import numpy as np
from scipy.integrate import odeint

# For finite difference estimates of derivatives
from mbam.finite_difference import jacobian_func, Avv_func, CD4, AvvCD4

exp = np.exp
# Time points to sample model. We do not observe t = 0, but is necessary for
# the ODE solver
ts = np.array([0.0, 1.0, 2.0, 5.0])
M = len(ts) - 1  # Number of predictions that the model makes
N = 2  # Target number of parameters


def rhs(y, t, x):
    return -exp(x[0]) * y / (exp(x[1]) + y)


# Returns our observation vector.
def r(x):
    return odeint(rhs, [1.0], ts, (x,))[1:, 0]


# Sensitivities ODE
def drhs(y, t, x):
    # y[0] = y, y[1] = dy/dx[0], y[2] = dy/dx[1]

    # deriviatve of rhs with respect to y
    drhsdy = -exp(x[0] + x[1]) / (exp(x[1]) + y[0]) ** 2
    # deriviatve of rhs with respect to x[0]
    drhsdx0 = -exp(x[0]) * y[0] / (exp(x[1]) + y[0])
    # deriviatve of rhs with respect to x[1]
    drhsdx1 = exp(x[0] + x[1]) * y[0] / (exp(x[1]) + y[0]) ** 2
    return [
        -exp(x[0]) * y[0] / (exp(x[1]) + y[0]),
        drhsdy * y[1] + drhsdx0,
        drhsdy * y[2] + drhsdx1,
    ]


# Jacobian
def j(x):
    return odeint(drhs, [1.0, 0.0, 0.0], ts, (x,))[1:, 1:]


# Alternatively, calculate jacobian using finite differences. See useful,
# higher-order formulas in `finite_difference.py`.
j_FD = jacobian_func(r, M, N, CD4, h=1e-2)


# Directional second derivative. This can also be done by either solving
# the sensitivity equations or using finite differences.  Here, we use finite
# differences.
Avv = Avv_func(r, AvvCD4, h=1e-2)



import numpy as np
from scipy.integrate import cumulative_trapezoid
from mbam import Geodesic
from mbam.finite_difference import jacobian_func, Avv_func

import model_linear as mlin
import model_exp as mexp


# Derivative functions for the linear model
J_fn = jacobian_func(mlin.f, mlin.M, mlin.N)
Avv_fn = Avv_func(mlin.f)

# Calculate the geodesic
# Linear model
geo_lin = Geodesic(mlin.f, J_fn, Avv_fn, mlin.xi, mlin.vi)
geo_lin.integrate(1000.0)
len_ts = len(geo_lin.ts)

# Exponential model with parameterspacenorm=True
geo_exp1 = Geodesic(
    mexp.f, mexp.jac_an, mexp.Avv_an, mexp.xi, mexp.vi, parameterspacenorm=True
)
geo_exp1.integrate(1000.0)
# Exponential model with parameterspacenorm=False
geo_exp2 = Geodesic(
    mexp.f, mexp.jac_an, mexp.Avv_an, mexp.xi, mexp.vi, parameterspacenorm=False
)
geo_exp2.integrate(1000.0)


def test_attributes():
    assert hasattr(geo_lin, "xs")
    assert hasattr(geo_lin, "rs")
    assert hasattr(geo_lin, "vs")
    assert hasattr(geo_lin, "ts")
    assert hasattr(geo_lin, "taus")


def test_derivative_shape():
    assert J_fn(mlin.xi).shape == (mlin.M, mlin.N)
    assert Avv_fn(mlin.xi, mlin.vi).shape == (mlin.M,)


def test_derivative_value():
    assert np.allclose(J_fn(mlin.xi), mlin.J)
    assert np.allclose(Avv_fn(mlin.xi, mlin.vi), 0.0)


def test_vs():
    """For a linear model, the velocitY in parameter space should be constant,
    and it should be the same as the initial velocity.
    """
    assert geo_lin.vs.shape == (len_ts, mlin.N)
    for vs in geo_lin.vs:
        assert np.allclose(vs, mlin.vi)


def test_xs():
    """Since vs is constant, then in each integration step, dx = vi * dt."""
    assert geo_lin.xs.shape == (len_ts, mlin.N)
    for ii in range(len_ts - 1):
        dx = mlin.vi * (geo_lin.ts[ii + 1] - geo_lin.ts[ii])
        assert np.allclose(dx, (geo_lin.xs[ii + 1] - geo_lin.xs[ii]))


def test_rs():
    """To test geo.rs values, we only need to check if geo.rs == J @ geo.xs."""
    assert geo_lin.rs.shape == (len_ts, mlin.M)
    for xs, rs in zip(geo_lin.xs, geo_lin.rs):
        assert np.allclose(rs, mlin.J @ xs)


def test_taus():
    """The product J.v gives dtau/dt. So, we can integrate this quantity and
    compare it to geo.taus.
    """
    assert geo_lin.taus.shape == (len_ts,)
    dtaus = [
        np.linalg.norm(J_fn(xs) @ vs) for xs, vs in zip(geo_lin.xs, geo_lin.vs)
    ]
    taus = np.append(0.0, cumulative_trapezoid(dtaus, geo_lin.ts))
    assert np.allclose(geo_lin.taus, taus)


def test_parameterspacenorm():
    """Test if the the implementation to use constant parameter space norm is
    implemented correctly. This is done by checking the geodesic velocity in
    parameter space. If parameterspacenorm=True, then the geodesic velocity
    should be constant.
    """
    # parameterspacenorm=True
    # To check if the velocity is constant we can compare the speed (the norm
    # of velocity) in each step to the mean speed value over the entire steps.
    # If the velocity is constant, the speed in each step should be close
    # enough to the mean speed, where the deviation is probably from some
    # numerical error.
    geo_exp1_speed = np.linalg.norm(geo_exp1.vs, axis=1)
    geo_exp1_speed_mean = np.mean(geo_exp1_speed)
    assert np.allclose(
        geo_exp1_speed, geo_exp1_speed_mean, atol=1e-4, rtol=1e-4
    )

    # parameterspacenorm=False
    # On the other side, if parameterspacenorm=False, then the velocity in
    # parameter space is not constant. The deviation of the speed from the mean
    # speed should be large.
    geo_exp2_speed = np.linalg.norm(geo_exp2.vs, axis=1)
    geo_exp2_speed_mean = np.mean(geo_exp2_speed)
    assert not np.allclose(geo_exp2_speed, geo_exp2_speed_mean)


if __name__ == "__main__":
    test_attributes()
    test_derivative_shape()
    test_derivative_value()
    test_vs()
    test_xs()
    test_rs()
    test_taus()
    test_parameterspacenorm()



import numpy as np


np.random.seed(1717)

# Dimensionality
M = 10
N = 5

# Time series
t = np.logspace(-4, 2, M)

# Define the model
J = np.array([t ** n for n in range(N)]).T


def f(x):
    return J @ x


# Initial conditions
xi = np.random.randn(N)
vi = np.random.randn(N)



from matplotlib import cm
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
from mbam import Geodesic, initial_velocity

from MMR import r, j, Avv


# Choose starting parameters
x = np.log([1.0, 1.0])
v = initial_velocity(x, j, Avv)


# Callback function used to monitor the geodesic after each step
def callback(g):
    # Integrate until the norm of the velocity has grown by a factor of 10
    # and print out some diagnotistic along the way
    print(
        "Iteration: %i, tau: %f, |v| = %f"
        % (len(g.vs), g.ts[-1], np.linalg.norm(g.vs[-1]))
    )
    return np.linalg.norm(g.vs[-1]) < 100.0


# Construct the geodesic
# It is usually not necessary to be very accurate here, so we set small
# tolerances
geo_forward = Geodesic(r, j, Avv, x, v, atol=1e-2, rtol=1e-2, callback=callback)

# Integrate
geo_forward.integrate(25.0)

# Plot the geodesic path to find the limit
# This should show the singularity at the "fold line" x[0] = x[1]
plt.figure()
plt.plot(geo_forward.ts, geo_forward.xs)
plt.xlabel("tau")
plt.ylabel("Parameter Values")
plt.show()


# Now do opposite direction

# Choose starting parameters
x = np.log([1.0, 1.0])
v = -initial_velocity(x, j, Avv)

# Construct the geodesic
# It is usually not necessary to be very accurate here, so we set small
# tolerances
geo_reverse = Geodesic(r, j, Avv, x, v, atol=1e-2, rtol=1e-2, callback=callback)

# Integrate
geo_reverse.integrate(25.0)

# Plot the geodesic path to find the limit
# This should show the singularity at the "fold line" x[0] = x[1]
# Add The geodesic to the same plot, with negative time points
plt.plot(-geo_reverse.ts, geo_reverse.xs)
plt.xlabel("tau")
plt.ylabel("Parameter Values")
plt.show()


# Now construct contour plots in parameter space and model manifold in data
# space

r0 = r([0.0, 0.0])
xs = np.linspace(-5, 5, 101)
X = np.empty((101, 101))
Y = np.empty((101, 101))
Z = np.empty((101, 101))
C = np.empty((101, 101))
for i, x in enumerate(xs):
    for j, y in enumerate(xs):
        temp = r([x, y])
        X[j, i], Y[j, i], Z[j, i] = temp
        C[j, i] = np.linalg.norm(temp - r0) ** 2

# Plot geodesic path in parameter space with cost contours
plt.figure()
plt.contourf(xs, xs, C, 50)
plt.plot(geo_forward.xs[:, 0], geo_forward.xs[:, 1], "r-")
plt.plot(geo_reverse.xs[:, 0], geo_reverse.xs[:, 1], "g-")
plt.plot([0], [0], "ro")
plt.xlim(-5, 5)
plt.ylim(-5, 5)
plt.show()

# Plot surface / geodesic in data space
plt.figure()
ax = plt.axes(projection="3d")
surf = ax.plot_surface(
    X, Y, Z, cmap=cm.coolwarm, linewidth=0, antialiased=False
)

geo = geo_forward
X = np.empty(len(geo.xs))
Y = np.empty(len(geo.xs))
Z = np.empty(len(geo.xs))
for i, x in enumerate(geo.xs):
    X[i], Y[i], Z[i] = r(x)
ax.plot(X, Y, Z, color="r", zorder=10)

geo = geo_reverse
X = np.empty(len(geo.xs))
Y = np.empty(len(geo.xs))
Z = np.empty(len(geo.xs))
for i, x in enumerate(geo.xs):
    X[i], Y[i], Z[i] = r(x)
ax.plot(X, Y, Z, color="g", zorder=10)


# Plot starting point of geodesic as a red dot
ax.scatter([r0[0]], [r0[1]], [r0[2]], c="r", s=25, zorder=10)
plt.show()



import numpy as np
import mbam.finite_difference as fd

from model_exp import M, N, f, jac_an, Avv_an, xi, vi


J = jac_an(xi)
Avv = Avv_an(xi, vi)


def test_jac_FD():
    J_fd = fd.jacobian_func(f, M, N, fd.FD, h=0.01)(xi)
    assert np.allclose(J_fd, J, rtol=1e-1), "Failed Jacobian FD"

    J_fd2 = fd.jacobian_func(f, M, N, fd.FD2, h=0.01)(xi)
    assert np.allclose(J_fd2, J, rtol=1e-1), "Failed Jacobian FD2"

    J_fd3 = fd.jacobian_func(f, M, N, fd.FD3, h=0.01)(xi)
    assert np.allclose(J_fd3, J, rtol=1e-1), "Failed Jacobian FD3"

    J_fd4 = fd.jacobian_func(f, M, N, fd.FD4, h=0.01)(xi)
    assert np.allclose(J_fd4, J, rtol=1e-1), "Failed Jacobian FD4"


def test_jac_CD():
    J_cd = fd.jacobian_func(f, M, N, fd.CD, h=0.01)(xi)
    assert np.allclose(J_cd, J, rtol=1e-1), "Failed Jacobian CD"

    J_cd4 = fd.jacobian_func(f, M, N, fd.CD4, h=0.01)(xi)
    assert np.allclose(J_cd4, J, rtol=1e-1), "Failed Jacobian CD4"


def test_Avv_CD():
    Avv_cd = fd.Avv_func(f, fd.AvvCD, h=0.01)(xi, vi)
    assert np.allclose(Avv_cd, Avv, rtol=1e-1), "Failed Avv CD"

    Avv_cd4 = fd.Avv_func(f, fd.AvvCD4, h=0.01)(xi, vi)
    assert np.allclose(Avv_cd4, Avv, rtol=1e-1), "Failed Avv CD4"


if __name__ == "__main__":
    test_jac_FD()
    test_jac_CD()
    test_Avv_CD()



"""A script containing utility functions: get sloppiest initial velocity.
"""


import numpy as np


def initial_velocity(x, jac, Avv, i=0, forward=True):
    """A routine for calculating the initial velocity.

    Parameters
    ----------
    x: (N,) np.ndarray
        Initial Parameter Values.
    jac: callable ``jac(x)``
        Function for calculating the jacobian.
    Avv: callable ``Avv(x, v)``
        Function for calculating a direction second derivative.
    i: int (optional)
        Index to choose which eigenvector direction to choose: 0 = sloppiest,
        1 = second sloppiest, etc.
    forward: bool (optional)
        A flag where if it is set to ``True``, the direction in which the
        velocity increases around ``x`` is chosen. Otherwise, the opposite
        direction is chosen.

    Returns
    -------
    v:
        Initial velocity vector.
    """

    j = jac(x)
    _, _, vh = np.linalg.svd(j)
    v = vh[-1 - i]
    a = -np.linalg.solve(j.T @ j, j.T @ Avv(x, v))
    # We choose the direction in which the velocity will increase, since this
    # is most likely to find a singularity quickest.
    if v @ a < 0:
        v *= -1

    if forward:
        return v / np.linalg.norm(v)
    else:
        return -v / np.linalg.norm(v)



#!/usr/bin/env python
import setuptools
import re

# List of dependecy packages
with open("requirements.txt") as f:
    install_requires = [line.strip() for line in f.readlines()]

# Find packages
packages = setuptools.find_packages(exclude=["examples"])

# Description of the package
description = "Model reduction by Manifold Boundary Approximation Method"
with open("README.md") as f:
    lines = f.readlines()
description = lines[1].strip()
long_description = "".join(lines)

# Get the current version number
with open("mbam/__init__.py") as fd:
    version = re.search('__version__ = "(.*)"', fd.read()).group(1)


setuptools.setup(
    name="mbam",
    version=version,
    author="Mark K. Transtrum",
    author_email="mktranstrum@byu.edu",
    maintainer="Yonatan Kurniawan",
    maintainer_email="kurniawanyo@outlook.com",
    url="https://github.com/mktranstrum/MBAM",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    packages=packages,
    classifiers=["Programming Language :: Python :: 3"],
    include_package_data=True,
    python_requires=">=3.6",
)



import numpy as np
from scipy.integrate import ode


# We construct a geodesic class that inherits from scipy's ode class
class Geodesic(ode):
    """A class to formulate and solve the geodesic equation of a model.

    Parameters
    ----------
    r: callable ``r(x)``
        Function for calculating the model. This is not needed for the
        geodesic, but is used to save the geodesic path in data space.
    j: callable ``j(x)``
        Function for calculating the jacobian of the model with respect to
        parameters. The default function uses forward difference method.
    Avv: callable ``Avv(x,v)``
        Function for calculating the second directional derivative of the
        model in a direction ``v``. The default function uses central
        difference method.
    x: (N,) np.ndarray
        Vector of initial parameter values.
    v: np.ndarray
        Vector of initial velocity.
    lam: float (optional)
        Set to non-zero to calculate geodesic on the model graph.
    dtd: (N, N) np.ndarray (optional)
        Metric for the parameter space contribution to the model graph.
    atol: float (optional)
        Absolute tolerance for solving the geodesic.
    rtol: float (optional)
        Relative tolerance for solving geodesic.
    callback: callable ``callback(geo)`` (optional)
        Function called after each geodesic step. Called as
        ``callback(geo)`` where ``geo`` is the current instance of the
        ``geodesic`` class.
    parameterspacenorm: bool (optional)
        Set to ``True`` to reparameterize the geodesic to have a constant
        parameter space norm and ``False`` to have a constant data space norm.
    invSVD: bool (optional)
        Set to true to use the singular value decomposition to calculate
        the inverse metric. This is slower, but can help with nearly
        singular FIM.

    Attributes
    ----------
    xs: (T, N) np.ndarray
        Geodesics in the parameter space. :math:`N` is the number of parameters.
    rs: (T, M) np.ndarray
        Geodesics in the data space. :math:`N` is the number of predictions.
    vs: (T, N) np.ndarray
        Velocity in the parameter space.
    ts: (T,) np.ndarray
        Geodesic arclength in the parameter space.
    taus: (T,) np.ndarray
        Geodesic arclength in the data space.

    Notes
    -----
    If the prediction points are weighted differently, then the weights need
    to be included in the Jacobian and directional second derivative
    functions. An option to do this is to use the residual function when
    computing the Jacobian and Avv, assuming that the weights are encoded
    in the residual.
    """

    def __init__(
        self,
        r,
        j,
        Avv,
        x,
        v,
        lam=0.0,
        dtd=None,
        atol=1e-6,
        rtol=1e-6,
        callback=None,
        parameterspacenorm=True,
        invSVD=False,
    ):

        self.N = len(x)

        # Functions needed in the calculation
        self.r, self.j, self.Avv = r, j, Avv

        # Initialize the ODE object
        ode.__init__(self, self.geodesic_rhs, jac=None)
        self.set_initial_value(x, v)
        ode.set_integrator(self, "vode", atol=atol, rtol=rtol)

        if callback is None:
            self.callback = callback_func
        else:
            self.callback = callback

        # Additional settings for RHS function
        self.lam = lam
        if dtd is None:
            self.dtd = np.eye(self.N)
        else:
            self.dtd = dtd
        self.parameterspacenorm = parameterspacenorm
        self.invSVD = invSVD

    def geodesic_rhs(self, t, xvtau):
        """This function implements the RHS of the geodesic equation. This
        equation is given by

        .. math::

            (J' J)^{-1} J' \mathbf{A}_{vv}

        This should not need to be edited or called directly by the user.

        Parameters
        ----------
        t: float
            "time" of the geodesic (tau).
        xvtau: (2N + 1,) np.ndarray
            Vector of current parameters, velocities, and data space distance.
        """
        x = xvtau[: self.N]
        v = xvtau[self.N : -1]
        v2 = v @ v
        if v2 == 0.0:
            return np.concatenate((v, 0, [0]))
        else:
            j = self.j(x)
            g = j.T @ j + self.lam * self.dtd
            Avv = self.Avv(x, v)
            if self.invSVD:
                u, s, vh = np.linalg.svd(j, 0)
                a = -vh.T @ (u.T @ Avv) / s
            else:
                a = -np.linalg.solve(g, j.T @ Avv)
            if self.parameterspacenorm:
                a -= a @ v * v / (v2)
            dtau = np.linalg.norm(j @ v)
            return np.concatenate((v, a, [dtau]))

    def set_initial_value(self, x, v):
        """Set the initial parameter values and velocity.

        Parameters
        ----------
        x: np.ndarray
            Vector of initial parameter values.
        v: np.ndarray
            Vector of initial velocity.
        """
        self.xs = np.array([x])
        self.vs = np.array([v])
        self.rs = np.array([self.r(x)])
        self.taus = np.array([0.0])
        self.ts = np.array([0.0])
        ode.set_initial_value(self, np.concatenate((x, v, [0.0])), 0.0)

    def step(self, dt=1.0):
        """Integrate the geodesic for one step.

        Parameters
        ----------
        dt: float
            target time step to use
        """
        ode.integrate(self, self.t + dt, step=1)
        self.xs = np.append(self.xs, [self.y[: self.N]], axis=0)
        self.vs = np.append(self.vs, [self.y[self.N : -1]], axis=0)
        self.rs = np.append(self.rs, [self.r(self.xs[-1])], axis=0)
        self.taus = np.append(self.taus, self.y[-1])
        self.ts = np.append(self.ts, self.t)

    def integrate(self, tmax, maxsteps=500):
        """Integrate the geodesic up to a fixed maximimum time (tau) or number
        of steps. After each step, the callback is called and the calculation
        will end if the callback returns False.

        Parameters
        ----------
        tmax: float
            Maximum time (tau) for integration
        maxsteps: int (optional)
            Maximum number of steps
        """
        cont = True
        while (
            self.successful()
            and len(self.xs) < maxsteps
            and self.t < tmax
            and cont
        ):
            self.step(tmax - self.t)
            cont = self.callback(self)


def callback_func(geo):
    """A default callback function. The geodesic can be halted by the callback
    returning False.

    Parameters
    ----------
    geo: object
        The current instance of the ``geodesic`` class.
    """
    return True



"""Contains functions to compute the derivative of the model with automatic
differentiation using JAX.
"""

import jax.numpy as jnp
from jax import jacfwd


def jacobian_func(f):
    """Generate a function to compute jacobian of a model. The output function
    will be compatible for :class:`~mbam.Geodesic`.

    Parameters
    ----------
    f: callable ``f(x)``
        Function to evaluate the model.

    Returns
    -------
    jacobian: callable ``jacobian(x)``
        A function that evaluates the Jacobian of the model ``f`` at parameter
        ``x``.

    Notes
    -----
    If the predictions are weighted differently, the argument ``f`` should be
    the residual function, assuming that the weights are encoded in the
    residual.
    """
    jacobian = jacfwd(f)
    return jacobian


def Avv_func(f):
    """Generate a function to compute directional second derivative Avv of a
    model. The output function will be compatible for :class:`~mbam.Geodesic`.

    Parameters
    ----------
    f: callable ``f(x)``
        Function to evaluate the model.

    Returns
    -------
    Avv: callable ``Avv(x, v)``
        A function that evaluates the directional second derivative (Avv) of
        the model ``f`` at parameter ``x`` in the ``v`` direction.

    Notes
    -----
    If the predictions are weighted differently, the argument ``f`` should be
    the residual function, assuming that the weights are encoded in the
    residual.
    """

    def Avv(x, v):
        def F(s):
            return f(x + v * s)

        return jacfwd(jacfwd(F))(0.0)

    return Avv



import pytest

import numpy as np

import model_exp as model
from model_exp import jac_an, Avv_an, xi, vi

if model.jax_avail:
    import mbam.auto_diff as ad

J = jac_an(xi)
Avv = Avv_an(xi, vi)


@pytest.mark.skipif(not model.jax_avail, reason="JAX not available")
def test_jac_AD():
    J_ad = ad.jacobian_func(model.f_ad)(xi)
    assert np.allclose(J_ad, J), "Failed Jacobian AD"


@pytest.mark.skipif(not model.jax_avail, reason="JAX not available")
def test_Avv_AD():
    Avv_ad = ad.Avv_func(model.f_ad)(xi, vi)
    assert np.allclose(Avv_ad, Avv), "Failed Avv AD"


if __name__ == "__main__":
    if model.jax_avail:
        test_jac_AD()
        test_Avv_AD()
